import sys
import math
from PyQt4 import QtCore, QtGui
from data import Node, Edge, Storm
from loader import StormLoader
from saver import StormSaver, SvgExporter
from saveoptions import Ui_dlgSaveOptions
from about import Ui_dlgAbout

# Brainstorm application for concept-mapping
# All code copyright activityworkshop.net, see license.txt file

# Class for GraphWidget in the middle of the gui
class GraphWidget(QtGui.QGraphicsView):
	def __init__(self):
		QtGui.QGraphicsView.__init__(self)

		self.timerId = 0
		self.drawing = False

		scene = Scene(self)
		scene.setItemIndexMethod(QtGui.QGraphicsScene.NoIndex)
		scene.setSceneRect(-350, -250, 700, 500)
		self.setScene(scene)
		self.setCacheMode(QtGui.QGraphicsView.CacheBackground)
		self.setRenderHint(QtGui.QPainter.Antialiasing)
		self.setTransformationAnchor(QtGui.QGraphicsView.AnchorUnderMouse)
		self.setResizeAnchor(QtGui.QGraphicsView.AnchorViewCenter)

		self.scale(0.8, 0.8)
		self.setMinimumSize(400, 400)

	def replaceScene(self, storm):
		self.storm = storm
		scene = self.scene()
		scene.clear()
		shouldrandomize = True
		for node in storm.nodes:
			node.graph = self
			scene.addItem(node)
			if node.pos().x() != 0 or node.pos().y() != 0:
				shouldrandomize = False
		for edge in storm.edges:
			scene.addItem(edge)
		# only randomize positions if they're all 0 (respect loaded positions)
		if shouldrandomize: self.randomizeNodePositions()
		self.calculateRanks()

	def updateScene(self, storm):
		# add unknown nodes, edges
		for node in storm.nodes:
			if (node not in self.scene().items()):
				node.graph = self
				self.scene().addItem(node)
		for edge in storm.edges:
			if (edge not in self.scene().items()):
				self.scene().addItem(edge)
		# need to subtract unwanted items too
		itemstodelete = []
		for item in self.scene().items():
			if item not in storm.nodes and item not in storm.edges:
				itemstodelete.append(item)
		for item in itemstodelete:
			self.scene().removeItem(item)
		self.calculateRanks()
		self.itemMoved()

	def itemMoved(self):
		if self.timerId == 0:
			self.timerId = self.startTimer(1000 / 25) # 25fps

	def keyPressEvent(self, event):
		key = event.key()

		if key == QtCore.Qt.Key_Plus:
			self.scaleView(1.2)
		elif key == QtCore.Qt.Key_Minus:
			self.scaleView(1 / 1.2)
		elif key == QtCore.Qt.Key_Space or key == QtCore.Qt.Key_Enter:
			self.randomizeNodePositions()
		else:
			QtGui.QGraphicsView.keyPressEvent(self, event)

	def randomizeNodePositions(self):
		for item in self.scene().items():
			if isinstance(item, Node):
				item.setPos(-150 + QtCore.qrand() % 300, -150 + QtCore.qrand() % 300)

	# Calculate the rank of each node iteratively
	def calculateRanks(self):
		nodes = [item for item in self.scene().items() if isinstance(item, Node)]
		edges = [item for item in self.scene().items() if isinstance(item, Edge)]
		# initialize with all ranks 1
		for node in nodes:
			node.rank = 1
			node.numEdges = 0
		for edge in edges:
			edge.source.numEdges += 1
			edge.dest.numEdges += 1
		DAMPING_FACTOR = 0.85
		keepGoing = True
		numLoops = 0
		while keepGoing and numLoops < 20:
			# reset nextrank for all nodes
			for node in nodes:
				node.nextrank = (1 - DAMPING_FACTOR)
			# feed rank through all edges
			for edge in edges:
				edge.source.nextrank += DAMPING_FACTOR * edge.dest.rank / edge.dest.numEdges
				edge.dest.nextrank += DAMPING_FACTOR * edge.source.rank / edge.source.numEdges
			# update all ranks
			keepGoing = False
			for node in nodes:
				if abs(node.rank - node.nextrank) > 0.01:
					keepGoing = True
				node.updateRank()
			numLoops += 1

	def timerEvent(self, event):
		if self.drawing: return
		self.drawing = True
		nodes = [item for item in self.scene().items() if isinstance(item, Node)]
		edges = [item for item in self.scene().items() if isinstance(item, Edge)]

		# attraction between two connected nodes
		for edge in edges:
			edge.source.calculateAttraction(edge.dest)

		# repulsion between every pair of nodes
		node1num = 0
		for node1 in nodes:
			node1num += 1
			node2num = 0
			for node2 in nodes:
				node2num += 1
				if node2 > node1:
					node1.calculateRepulsion(node2)

		itemsMoved = False
		for node in nodes:
			if node.advance():
				itemsMoved = True

		for edge in edges:
			edge.adjust()

		if not itemsMoved:
			self.killTimer(self.timerId)
			self.timerId = 0
		self.drawing = False

	def wheelEvent(self, event):
		self.scaleView(math.pow(2.0, -event.delta() / 240.0))

	def scaleView(self, scaleFactor):
		factor = self.matrix().scale(scaleFactor, scaleFactor).mapRect(QtCore.QRectF(0, 0, 1, 1)).width()
		if factor < 0.07 or factor > 100:
			return
		self.scale(scaleFactor, scaleFactor)

# Subclass for graphics scene
class Scene(QtGui.QGraphicsScene):
	def __init__(self, parent):
		QtGui.QGraphicsScene.__init__(self, parent)

	# override the mouse event handling to pass unprocessed events upwards
	def mousePressEvent(self, event):
		QtGui.QGraphicsScene.mousePressEvent(self, event)
		if event.button() == QtCore.Qt.RightButton and not event.isAccepted():
			self.parent().storm.processRightClick(event, self)


# Main class for Brainstorm
class Brainstorm(QtGui.QMainWindow):

	def __init__(self):
		QtGui.QMainWindow.__init__(self)
		self.setupUi()
		self.storm = None
		self.setStorm(Storm())

	# set the current storm and its callback functions
	def setStorm(self, storm):
		if (self.storm):
			self.storm.menuCallback = None
			self.storm.updateCallback = None
		self.storm = storm
		if (storm):
			storm.menuCallback = self.showContextMenu
			storm.updateCallback = self.updateScene
			self.gwidget.replaceScene(self.storm)

	# Initialise the user interface
	def setupUi(self):
		self.setObjectName("MainWindow")
		self.resize(551,343)
		self.gwidget = GraphWidget()
		self.setCentralWidget(self.gwidget)
		self.menubar = QtGui.QMenuBar(self)
		self.menubar.setGeometry(QtCore.QRect(0,0,551,26))
		self.menubar.setObjectName("menubar")
		self.menuFile = QtGui.QMenu(self.menubar)
		self.menuFile.setObjectName("menuFile")
		self.menuHelp = QtGui.QMenu(self.menubar)
		self.menuHelp.setObjectName("menuHelp")
		self.setMenuBar(self.menubar)
		self.statusbar = QtGui.QStatusBar(self)
		self.statusbar.setObjectName("statusbar")
		self.setStatusBar(self.statusbar)
		self.actionNew = QtGui.QAction(self)
		self.actionOpen = QtGui.QAction(self)
		self.actionSave = QtGui.QAction(self)
		self.actionSaveSvg = QtGui.QAction(self)
		self.actionAbout_Brainstorm = QtGui.QAction(self)
		self.actionAbout_Brainstorm.setObjectName("actionAbout_Brainstorm")
		self.menuFile.addAction(self.actionNew)
		self.menuFile.addAction(self.actionOpen)
		self.menuFile.addAction(self.actionSave)
		self.menuFile.addAction(self.actionSaveSvg)
		self.menuHelp.addAction(self.actionAbout_Brainstorm)
		self.menubar.addAction(self.menuFile.menuAction())
		self.menubar.addAction(self.menuHelp.menuAction())
		# connect slots
		self.connect(self.actionNew, QtCore.SIGNAL("triggered()"), self.newStorm)
		self.connect(self.actionOpen, QtCore.SIGNAL("triggered()"), self.openFile)
		self.connect(self.actionSave, QtCore.SIGNAL("triggered()"), self.saveStorm)
		self.connect(self.actionSaveSvg, QtCore.SIGNAL("triggered()"), self.saveSvg)
		self.connect(self.actionAbout_Brainstorm, QtCore.SIGNAL("triggered()"), self.showAbout)
		# texts
		self.setWindowTitle("Brainstorm")
		self.setStatusTip("")
		self.menuFile.setTitle("File")
		self.menuHelp.setTitle("Help")
		self.actionNew.setText("New")
		self.actionOpen.setText("Open")
		self.actionSave.setText("Save Storm")
		self.actionSaveSvg.setText("Save as SVG")
		self.actionAbout_Brainstorm.setText("About Brainstorm")

	def newStorm(self):
		self.setStorm(Storm())

	def openFile(self):
		filename = QtGui.QFileDialog.getOpenFileName(self, 'Open file')
		if filename == None or len(filename) == 0:
			return
		self.setStorm(StormLoader().loadFile(filename))

	def saveStorm(self):
		if (self.storm):
			# select file to save
			filename = QtGui.QFileDialog.getSaveFileName(self, 'Save file')
			if filename == None or len(filename) == 0:
				return
			saveDialog = SaveDialog(self)
			saveDialog.init(self.storm, filename)
			saveDialog.show()
		else:
			print "no storm to save!"

	def saveSvg(self):
		if (self.storm):
			filename = QtGui.QFileDialog.getSaveFileName(self, 'Save file')
			if filename == None or len(filename) == 0:
				return
			SvgExporter().exportSvg(self.storm, filename)
		else:
			print "no storm to save!"

	# show context menu for right-click on either node or edge
	def showContextMenu(self, event, item):
		menu = QtGui.QMenu(self)
		newNodeAction = connectAction = deleteNodeAction = deleteEdgeAction = newSoloNodeAction = None
		if isinstance(item, Node):
			newNodeAction = menu.addAction("New node attached to '%s'" % item.label)
			connectAction = menu.addAction("New connection to '%s'" % item.label)
			menu.addSeparator()
			deleteNodeAction = menu.addAction("Delete node '%s'" % item.label)
		elif isinstance(item, Edge):
			deleteEdgeAction = menu.addAction("Delete edge '%s'-'%s'" % (item.source.label, item.dest.label))
		elif isinstance(item, Scene):
			newSoloNodeAction = menu.addAction("Create new node")

		action = menu.exec_(event.screenPos())
		if (action == None): return
		if (action == newNodeAction):
			# create new node attached to current one
			self.createNewNode(event, item)
		elif action == newSoloNodeAction:
			self.createNewNode(event, None)
		elif (action == connectAction):
			# new connection from this node - just remember which node
			self.storm.createConnectionFrom(item)
		elif (action == deleteNodeAction):
			# delete current node
			self.storm.deleteNode(item)
			self.updateScene()
		elif (action == deleteEdgeAction):
			# delete current edge
			self.storm.deleteEdge(item)
			self.updateScene()

	# create new node optionally connected to an existing one
	def createNewNode(self, event, currnode):
		# prompt for label
		ans, ok = QtGui.QInputDialog.getText(self, 'Create New Node', 'Node label:', QtGui.QLineEdit.Normal)
		if ok:
			# ask storm for unused nodeid
			newid = self.storm.getUnusedNodeId()
			newnode = Node(None, newid, str(ans))
			if currnode:
				newnode.setPos(currnode.pos().x() - 12, currnode.pos().y() + 12)
			else: newnode.setPos(event.pos())
			self.storm.addNode(newnode)
			newnode.storm = self.storm
			# connect
			if currnode: self.storm.addEdge(currnode.nodeid, newid)
			self.updateScene()

	def updateScene(self):
		self.gwidget.updateScene(self.storm)

	# Show the about dialog
	def showAbout(self):
		aboutDialog = AboutDialog(self)
		aboutDialog.show()


# Class for Save options Dialog
class SaveDialog(QtGui.QDialog, Ui_dlgSaveOptions):
	def __init__(self, parent):
		QtGui.QDialog.__init__(self, parent)
		Ui_dlgSaveOptions.__init__(self)
		self.setupUi(self)
	# reinitialise dialog with storm information and filename to save
	def init(self, storm, filename):
		self.storm = storm
		self.outputfilename = filename
		self.txtName.setText(storm.title)
		self.txtAuthor.setText(storm.author)
		self.txtName.setFocus()
	def accept(self):
		# modify storm to new title / author
		self.storm.title = self.txtName.text()
		self.storm.author = self.txtAuthor.text()
		StormSaver().saveStorm(self.storm, self.outputfilename, self.chkSavePositions.isChecked())
		self.close()

# Class for About Dialog
class AboutDialog(QtGui.QDialog, Ui_dlgAbout):
	def __init__(self, parent):
		QtGui.QDialog.__init__(self, parent)
		Ui_dlgAbout.__init__(self)
		self.setupUi(self)


if __name__ == "__main__":
	app = QtGui.QApplication(sys.argv)
	QtCore.qsrand(QtCore.QTime(0,0,0).secsTo(QtCore.QTime.currentTime()))

	bs = Brainstorm()
	bs.show()
	sys.exit(app.exec_())

