from PyQt4 import QtCore, QtGui

# Brainstorm application for concept-mapping
# All code copyright activityworkshop.net, see license.txt file

# Class for each edge in the graph
class Edge(QtGui.QGraphicsItem):

	def __init__(self, sourceNode, destNode):
		QtGui.QGraphicsItem.__init__(self)

		self.sourcePoint = QtCore.QPointF()
		self.destPoint = QtCore.QPointF()
		self.setAcceptedMouseButtons(QtCore.Qt.RightButton)
		self.source = sourceNode
		self.dest = destNode
		self.adjust()

	def sourceNode(self):
		return self.source

	def setSourceNode(self, node):
		self.source = node
		self.adjust()

	def destNode(self):
		return self.dest

	def setDestNode(self, node):
		self.dest = node
		self.adjust()

	def adjust(self):
		if not self.source or not self.dest:
			return

		line = QtCore.QLineF(self.mapFromItem(self.source, self.source.boundingRect().center()), self.mapFromItem(self.dest, self.dest.boundingRect().center()))
		if line.length() == 0.0:
			return

		self.prepareGeometryChange()
		self.sourcePoint = line.p1()
		self.destPoint = line.p2()

	# define rectangle enclosing drawable area
	def boundingRect(self):
		if not self.source or not self.dest:
			return QtCore.QRectF()
		extra = 5
		return QtCore.QRectF(self.sourcePoint,
			QtCore.QSizeF(self.destPoint.x() - self.sourcePoint.x(),
			self.destPoint.y() - self.sourcePoint.y())).normalized().adjusted(-extra, -extra, extra, extra)

	# define shape for mouse click detection
	def shape(self):
		path = QtGui.QPainterPath();
		path.moveTo(self.sourcePoint)
		path.lineTo(self.destPoint)
		stroker = QtGui.QPainterPathStroker()
		stroker.setWidth(10.0)
		return stroker.createStroke(path);

	def paint(self, painter, option, widget):
		if not self.source or not self.dest:
			return
		# Draw the line itself.
		line = QtCore.QLineF(self.sourcePoint, self.destPoint)
		if line.length() == 0.0:
			return
		painter.setPen(QtGui.QPen(QtCore.Qt.black, 1, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin))
		painter.drawLine(line)

	def mousePressEvent(self, event):
		if event.button() == QtCore.Qt.RightButton:
			# pass event on up to the storm
			self.storm.processRightClick(event, self)


# Class for each node
class Node(QtGui.QGraphicsTextItem):
	ATTRACTION_FACTOR = 0.25
	REPULSION_FACTOR = 350
	REPULSION_RANGE = 300
	DRAG_FACTOR = 0.8
	LOW_VELOCITY = 2
	SMALLEST_VELOCITY = 0.5

	def __init__(self, graphWidget, nodeid, label, x=0, y=0):
		QtGui.QGraphicsTextItem.__init__(self, " " + label + " ")
		self.graph = graphWidget
		self.label = label
		self.setFlag(QtGui.QGraphicsItem.ItemIsMovable)
		try:
			self.setFlag(QtGui.QGraphicsItem.ItemSendsGeometryChanges) # required from version 4.6
		except: pass # version 4.5 throws error
		self.setPos(float(x), float(y))
		self.velx = 0
		self.vely = 0
		self.nodeid = nodeid
		self.setZValue(1)
		self.defaultPointSize = self.font().pointSize()
		self.numEdges = 0

	def calculateRepulsion(self, other):
		# calculate distance between points
		line = QtCore.QLineF(self.mapFromItem(other, other.boundingRect().center()), self.boundingRect().center())
		edgelen = line.length()
		if edgelen > 0 and edgelen < Node.REPULSION_RANGE:
			push = Node.REPULSION_FACTOR / edgelen / edgelen
			velxIncrement = line.dx() * push
			velyIncrement = line.dy() * push
			self.addVelocity(velxIncrement, velyIncrement)
			other.addVelocity(-velxIncrement, -velyIncrement)

	def calculateAttraction(self, other):
		# calculate distance between points
		line = QtCore.QLineF(self.mapFromItem(other, other.boundingRect().center()), self.boundingRect().center())
		if line.length() > 0:
			velxIncrement = line.dx() * Node.ATTRACTION_FACTOR
			velyIncrement = line.dy() * Node.ATTRACTION_FACTOR
			self.addVelocity(-velxIncrement, -velyIncrement)
			other.addVelocity(velxIncrement, velyIncrement)

	def addVelocity(self, xInc, yInc):
		self.velx += xInc
		self.vely += yInc

	def advance(self):
		if not self.scene() or self.scene().mouseGrabberItem() is self:
			self.velx = self.vely = 0
			return False
		# slow things down
		self.velx *= Node.DRAG_FACTOR
		self.vely *= Node.DRAG_FACTOR
		# slow down the slow things more
		if abs(self.velx) < Node.LOW_VELOCITY: self.velx *= 0.95
		if abs(self.vely) < Node.LOW_VELOCITY: self.vely *= 0.95
		if abs(self.velx) < Node.SMALLEST_VELOCITY and abs(self.vely) < Node.SMALLEST_VELOCITY:
			self.velx = self.vely = 0.0
			return False
		# Use velx, vely to calculate new position
		sceneRect = self.scene().sceneRect()
		newPos = self.pos() + QtCore.QPointF(self.velx, self.vely)
		newPos.setX(min(max(newPos.x(), sceneRect.left() + 10), sceneRect.right() - 10))
		newPos.setY(min(max(newPos.y(), sceneRect.top() + 10), sceneRect.bottom() - 10))
		self.setPos(newPos)
		return True

	# update the rank and change the appearance accordingly
	def updateRank(self):
		self.rank = self.nextrank
		font = self.font()
		font.setBold(self.rank > 1.2)
		if self.rank < 0.9:
			font.setPointSize(self.defaultPointSize - 3)
		elif self.rank > 1.25:
			font.setPointSize(self.defaultPointSize + 3)
		else:
			font.setPointSize(self.defaultPointSize)
		self.setFont(font)

	def paint(self, painter, option, widget):
		painter.setBrush(QtCore.Qt.white)
		painter.drawEllipse(self.boundingRect())
		QtGui.QGraphicsTextItem.paint(self, painter, option, widget)

	# define shape for mouse click detection
	def shape(self):
		path = QtGui.QPainterPath();
		path.addEllipse(self.boundingRect())
		return path

	def itemChange(self, change, value):
		if change == QtGui.QGraphicsItem.ItemPositionChange:
			if self.graph != None:
				self.graph.itemMoved()
		return QtGui.QGraphicsItem.itemChange(self, change, value)

	def mousePressEvent(self, event):
		if event.button() == QtCore.Qt.RightButton:
			return
		self.update()
		if self.storm.connectFromNode:
			self.storm.addEdge(self.storm.connectFromNode.nodeid, self.nodeid)
			self.storm.updateScene()
		try:
			super(Node, self).mousePressEvent(event) #4.6
		except:
			QtGui.QGraphicsItem.mousePressEvent(self, event) #4.5

	def mouseReleaseEvent(self, event):
		if event.button() == QtCore.Qt.RightButton:
			# pass event on up to the storm
			self.storm.processRightClick(event, self)
		self.update()
		try:
			super(Node, self).mouseReleaseEvent(event) #4.6
		except:
			QtGui.QGraphicsItem.mouseReleaseEvent(self, event) #4.5

	def mouseDoubleClickEvent(self, event):
		# prompt for label (shouldn't be responsibility of data object!)
		ans, ok = QtGui.QInputDialog.getText(None, 'Edit Node', 'Node label:', QtGui.QLineEdit.Normal, self.label)
		if ok:
			self.label = str(ans)
			self.setPlainText(" " + self.label + " ")
			self.storm.updateScene()



# Class to hold a set of nodes and their connections
class Storm:
	def __init__(self):
		self.title = ""
		self.author = ""
		self.nodes = []
		self.edges = []
		self.nodedict = {}
		self.connectFromNode = None

	def addNode(self, node):
		if self.nodedict.get(node.nodeid):
			print "node ", node.nodeid, "already exists"
		else:
			self.nodedict[node.nodeid] = node
			self.nodes.append(node)
		node.storm = self
		self.connectFromNode = None

	def addEdge(self, sourceid, destid):
		sourceNode = self.nodedict.get(sourceid)
		destNode = self.nodedict.get(destid)
		if sourceNode == None or destNode == None:
			print "Can't find nodes for edge", sourceid, "-", destid
		elif sourceNode == destNode:
			print "Can't link node to itself"
		else:
			for edge in self.edges:
				if (edge.source.nodeid == sourceid and edge.dest.nodeid == destid) or \
				 (edge.dest.nodeid == sourceid and edge.source.nodeid == destid):
					print "edge already exists"
					return
			edge = Edge(sourceNode, destNode)
			self.edges.append(edge)
			edge.storm = self
		self.connectFromNode = None

	def processRightClick(self, event, item):
		self.connectFromNode = None
		self.menuCallback(event, item)

	def getUnusedNodeId(self):
		nodenum = len(self.nodes) + 1
		while self.nodedict.get("node" + str(nodenum)):
			nodenum += 1
		return "node" + str(nodenum)

	def deleteNode(self, node):
		if self.nodedict.get(node.nodeid):
			del self.nodedict[node.nodeid]
			self.nodes.remove(node)
			# also remove all connected edges
			newedges = []
			for edge in self.edges:
				if edge.source != node and edge.dest != node:
					newedges.append(edge)
			self.edges = newedges
		self.connectFromNode = None

	def deleteEdge(self, edge):
		self.edges.remove(edge)
		self.connectFromNode = None

	def createConnectionFrom(self, node):
		self.connectFromNode = node

	def updateScene(self):
		self.updateCallback()

